/*
   Point Sprites Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to display point sprites in Direct3D.
   When point sprites are set up, all we have to do is render a bunch
   of vertex points and Direct3D will take over.
*/


// Demo: Point Sprites Demo.


#define WIN32_LEAN_AND_MEAN
#define VC_LEANMEAN

#include<d3dx9.h>

#define D3DFVF_D3DVertex (D3DFVF_XYZ | D3DFVF_DIFFUSE)

inline unsigned long FtoDW(float val) { return *((unsigned long*)&val); }


// Function Prototypes...
bool InitializeDirect3D(HWND hwnd, bool fullscreen);
bool InitializeObject();
void RenderScene();
void SetPerpective();
void ShutdownDirect3D();


// Here are our Direct3D objects.
LPDIRECT3D9 Direct3D_Object = NULL;
LPDIRECT3DDEVICE9 D3D_Device = NULL;
LPDIRECT3DVERTEXBUFFER9 Vertex_Buffer = NULL;
LPDIRECT3DTEXTURE9 Texture = NULL;

struct PointSprite
{
   float x, y, z;
   DWORD color;
};


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch(message)
       {
		   case WM_KEYUP:
            if(wParam == VK_ESCAPE)
               PostQuitMessage(0);
            break;

         case WM_DESTROY:
         case WM_CLOSE:
            PostQuitMessage(0);
            break;

            default:
               break;
       }

    return DefWindowProc( hwnd, message, wParam, lParam );
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                           // Message object.
   HWND hwnd;                                         // Handle to the window.
   WNDCLASSEX windowClass;                            // Window class object.
   bool done = false;                                 // true then we should exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);           // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;       // style of the window.
   windowClass.lpfnWndProc = WndProc;                 // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                        // Extra class information.
   windowClass.cbWndExtra = 0;                        // Extra window information.
   windowClass.hInstance = hInstance;                 // Handle of application Instance.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);// Window Icon.
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW); // Load mouse cursor.
   windowClass.hbrBackground = NULL;                  // Background color.
   windowClass.lpszMenuName = NULL;                   // Menu.
   windowClass.lpszClassName = "UGPClass";            // Name of the window class.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);// Minimized window icon.

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // After your class has been registered then you are ready to create your window.
   hwnd = CreateWindowEx(NULL,                                    // The extended style.
                         "UGPClass",                              // Window class.
                         "Point Sprites - by The Programming Ace",// Window name.
                          WS_OVERLAPPEDWINDOW | WS_VISIBLE |      // Window style.
                          WS_SYSMENU |WS_CLIPCHILDREN |           // Window style.
                          WS_CLIPSIBLINGS,                        // Window style.
                          100, 100,                               // X, Y coords.
                          640, 480,                               // Window size.
                          NULL,                                   // Handle to parent window.
                          NULL,                                   // Menu.
                          hInstance,                              // Handle to app instance.
                          NULL);                                  // Pointer to window.

   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW);    // Show the window.
   UpdateWindow(hwnd);           // Update its display.

   done = false;                 // false = run program, true means stop.

   // Initialize Direct3D.  If fails then we don't want to run the application.
   if(!InitializeDirect3D(hwnd, false)) done = true;

   // Application loop.
   while(!done)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     done = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
            }
      }


    // Here we shutdown Direct 3D.
	ShutdownDirect3D();

   // Here we unregister the window class with the OS.
   UnregisterClass("UGPClass", windowClass.hInstance);

   return (int)msg.wParam;
}


bool InitializeDirect3D(HWND hwnd, bool fullscreen)
{
   // This object will allow us to set the display mode of the screen.
   D3DDISPLAYMODE DisplayMode;

   // This will allow us to set the parameters of the screen.
   D3DPRESENT_PARAMETERS Present_Parameters;

   // This is used to get the capabilities of the hardware.
   D3DCAPS9 D3DCaps;

   // It is always a good idea to clear out memory in object although not necessary.
   ZeroMemory(&Present_Parameters, sizeof(Present_Parameters));

   // Create the Direct3D object to get everything started.
   Direct3D_Object = Direct3DCreate9(D3D_SDK_VERSION);

   // Error checking.  Make sure that it was successful.
   if(Direct3D_Object == NULL) return false;

   // This function will get the display mode of the device and place it in DisplayMode.
   if(FAILED(Direct3D_Object->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &DisplayMode)))
      return false;

   // Get the capabilities of the hardware.
	if(FAILED(Direct3D_Object->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &D3DCaps)))
		return false;

   // Test which is supported, hardware or software vertex processing.
	DWORD VertexProcessing = 0;

	if(D3DCaps.VertexProcessingCaps != 0)
		VertexProcessing |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else
		VertexProcessing |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;

   // Here we are setting the applications parameters...
   if(fullscreen)
      {
         Present_Parameters.Windowed = FALSE;               // Window mode (fullscreen).
         Present_Parameters.BackBufferWidth = DisplayMode.Width;
         Present_Parameters.BackBufferHeight = DisplayMode.Height;
      }
   else
      Present_Parameters.Windowed = TRUE;                   // Window mode (not fullscreen).
   Present_Parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;   // Dealing with animation (see doc).
   Present_Parameters.BackBufferFormat = DisplayMode.Format;// Render to the area of the screen.
   Present_Parameters.BackBufferCount = 1;                  // Number of back buffers.
   Present_Parameters.EnableAutoDepthStencil = TRUE;        // Check documentation.
   Present_Parameters.AutoDepthStencilFormat = D3DFMT_D16;  // Check documentation.

   // Now we must create the rendering device.
   if(FAILED(Direct3D_Object->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hwnd,
                                 VertexProcessing, &Present_Parameters, &D3D_Device)))
      return false;

   // One last check to be sure.
   if(D3D_Device == NULL) return false;

   D3D_Device->SetRenderState(D3DRS_LIGHTING, FALSE);
	D3D_Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
   D3D_Device->SetRenderState(D3DRS_ZENABLE, FALSE);

   InitializeObject();

   // Set the texture min/mag info.
   D3D_Device->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	D3D_Device->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

   // Set the window perpective.
   SetPerpective();

   return true;
}


bool InitializeObject()
{
	PointSprite sprites[4];
   sprites[0].x = -0.5f; sprites[0].y = -0.5f; sprites[0].z = 2.0f;
   sprites[0].color = D3DCOLOR_XRGB(255, 0, 255);

   sprites[1].x = 0.5f; sprites[1].y = -0.5f; sprites[1].z = 2.0f;
   sprites[1].color = D3DCOLOR_XRGB(0, 255, 0);

   sprites[2].x = 0.5f; sprites[2].y = 0.5f; sprites[2].z = 2.0f;
   sprites[2].color = D3DCOLOR_XRGB(0, 0, 255);

   sprites[3].x = -0.5f; sprites[3].y = 0.5f; sprites[3].z = 2.0f;
   sprites[3].color = D3DCOLOR_XRGB(255, 0, 0);

   // Create the vertex buffer that will hold the triangle.
	if(FAILED(D3D_Device->CreateVertexBuffer(sizeof(sprites), D3DUSAGE_DYNAMIC |
                           D3DUSAGE_WRITEONLY | D3DUSAGE_POINTS, D3DFVF_D3DVertex,
                           D3DPOOL_DEFAULT, &Vertex_Buffer, NULL))) return false;

   PointSprite *ptr;

   // Copy data into the buffer.
   if(FAILED(Vertex_Buffer->Lock(0, sizeof(sprites), (void**)&ptr, 0)))
      return false;

   memcpy(ptr, sprites, sizeof(sprites));
   Vertex_Buffer->Unlock();

   // Load the texture image.
   if(FAILED(D3DXCreateTextureFromFile(D3D_Device, "sprite.tga", &Texture)))
      return false;

   return true;
}


void SetPerpective()
{
   // Projection.
   D3DXMATRIXA16 Projection;

   // Set the perspective information.
   D3DXMatrixPerspectiveFovLH(&Projection, 45.0f, 640/480, 0.1f, 500.0f);

   // Set the projection.
   D3D_Device->SetTransform(D3DTS_PROJECTION, &Projection);
}


void RenderScene()
{
   // Clear the screen to black.
   D3D_Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                     D3DCOLOR_XRGB(0,0,0), 1.0f, 0);
    
   D3D_Device->BeginScene();

      D3D_Device->SetTexture(0, Texture);

      D3D_Device->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
      D3D_Device->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);

      D3D_Device->SetRenderState(D3DRS_POINTSPRITEENABLE, TRUE);
      D3D_Device->SetRenderState(D3DRS_POINTSCALEENABLE, TRUE);
      D3D_Device->SetRenderState(D3DRS_POINTSIZE, FtoDW(1.0));
      D3D_Device->SetRenderState(D3DRS_POINTSIZE_MIN, FtoDW(1.0f));
      D3D_Device->SetRenderState(D3DRS_POINTSCALE_A, FtoDW(0.0f));
      D3D_Device->SetRenderState(D3DRS_POINTSCALE_B, FtoDW(0.0f));
      D3D_Device->SetRenderState(D3DRS_POINTSCALE_C, FtoDW(1.0f));

      D3D_Device->SetStreamSource(0, Vertex_Buffer, 0, sizeof(PointSprite));
      D3D_Device->SetFVF(D3DFVF_D3DVertex);
      D3D_Device->DrawPrimitive(D3DPT_POINTLIST, 0, 4);

      D3D_Device->SetRenderState(D3DRS_POINTSPRITEENABLE, FALSE);
      D3D_Device->SetRenderState(D3DRS_POINTSCALEENABLE, FALSE);
      D3D_Device->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);

   D3D_Device->EndScene();
    
   // Swap the old frame with the new one.
   D3D_Device->Present(NULL, NULL, NULL, NULL);
}


void ShutdownDirect3D()
{
   // Here we release the Direct3D objects.
   if(D3D_Device != NULL)
      {
         D3D_Device->Release();
         D3D_Device = NULL;
      }

   if(Direct3D_Object != NULL)
      {
         Direct3D_Object->Release();
         Direct3D_Object = NULL;
      }
   
   if(Texture != NULL)
      {
         Texture->Release();
         Texture = NULL;
      }

   if(Vertex_Buffer != NULL)
      {
         Vertex_Buffer->Release();
         Vertex_Buffer = NULL;
      }
}


// Recap:

// Point sprites are very useful, especially when it comes to
// particle systems and things that are small and many.  Once we
// call a few api functions all we need to do is render a vertex point
// for every particle we want to display.


// Copyright October 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com